#!/usr/bin/env node

'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

require('core-js/modules/es.array.iterator');
var _reduce = _interopDefault(require('ramda/src/reduce'));
var program = _interopDefault(require('commander'));
require('core-js/modules/es.string.replace');
var _map = _interopDefault(require('ramda/src/map'));
var _prop = _interopDefault(require('ramda/src/prop'));
var _descend = _interopDefault(require('ramda/src/descend'));
var _sort = _interopDefault(require('ramda/src/sort'));
var _lensProp = _interopDefault(require('ramda/src/lensProp'));
var _set = _interopDefault(require('ramda/src/set'));
var _compose = _interopDefault(require('ramda/src/compose'));
var _multiply = _interopDefault(require('ramda/src/multiply'));
var _always = _interopDefault(require('ramda/src/always'));
var _ifElse = _interopDefault(require('ramda/src/ifElse'));
var _divide = _interopDefault(require('ramda/src/divide'));
var _flip = _interopDefault(require('ramda/src/flip'));
var _any = _interopDefault(require('ramda/src/any'));
var _filter = _interopDefault(require('ramda/src/filter'));
var _find = _interopDefault(require('ramda/src/find'));
var _endsWith = _interopDefault(require('ramda/src/endsWith'));
var _startsWith = _interopDefault(require('ramda/src/startsWith'));
var _anyPass = _interopDefault(require('ramda/src/anyPass'));
var _replace = _interopDefault(require('ramda/src/replace'));
var _is = _interopDefault(require('ramda/src/is'));
var snippets = _interopDefault(require('../lib/snippets.json'));
var _pick = _interopDefault(require('ramda/src/pick'));
var _toPairs = _interopDefault(require('ramda/src/toPairs'));
var _join = _interopDefault(require('ramda/src/join'));
var Renderer = _interopDefault(require('marked-terminal'));
var chalk = _interopDefault(require('chalk'));
var marked = _interopDefault(require('marked'));
var cliHighlight = require('cli-highlight');
var clipboardy = require('clipboardy');

const enforceSingleNewLine = x => _is(String, x) ? _compose(_replace(/[\r\n]+/g, '\n'), _replace(/[\r\n]*$/, ''))(x) : x;
const randomNumberInRange = (min, max) => Math.floor(Math.random() * (max - min + 1)) + min;
const startsOrEndsWith = (val, query) => _anyPass([_startsWith, _endsWith])(query, val);

const fuzzyMatch = query => x => new RegExp(query).test(x.id);

const getSnippet = (snippets, id) => _find(x => x.id === id, snippets);
const getSnippetsByTag = (snippets, id) => _filter(x => _any(y => y === id, x.tags), snippets);
const randomSnippet = snippets => snippets[randomNumberInRange(0, snippets.length - 1)];

const computeRelevance = (id, query) => {
  const base = query.length * 10;
  const diffInChars = id.replace(query, '').length + 1;

  const divideByDiffInChars = _flip(_divide)(diffInChars);

  const multi = _ifElse(_always(startsOrEndsWith(id, query)), _multiply(0.75), _multiply(0.5));

  if (id === query) return base;
  return _compose(divideByDiffInChars, multi)(base);
};

const setSearchRelevance = query => x => _set(_lensProp('relevance'), computeRelevance(x.id, query), x);

const searchSnippets = (snippets, query) => _compose(_sort(_descend(_prop('relevance'))), _map(setSearchRelevance(query)), _filter(fuzzyMatch(query)))(snippets);

marked.setOptions({
  renderer: new Renderer()
});
const colorizedPrint = x => {
  const printMap = {
    code: y => cliHighlight.highlight(y, {
      language: 'javascript'
    }),
    description: y => enforceSingleNewLine(marked(y)),
    example: y => cliHighlight.highlight(y, {
      language: 'javascript'
    }),
    id: y => chalk.magenta.bold(y),
    tags: _compose(chalk.gray.italic, _join(', '))
  };

  const print = y => _compose(_replace(/\n$/, ''), _join('\n'), _map(([k, v]) => printMap[k](v) + (k === 'id' ? '' : '\n')), _toPairs)(y);

  return _compose(_join('\n\n'), _map(print))(x);
};
const printSnippet = ({
  cp,
  layout,
  json
}, x) => {
  const arr = _is(Array, x) ? x : [x];
  const layoutMap = {
    c: 'code',
    d: 'description',
    e: 'example',
    i: 'id',
    t: 'tags'
  };

  const keysByLayout = _map(k => layoutMap[k])(Array.from(layout || 'itced'));

  if (x == null) return '';

  if (cp) {
    clipboardy.writeSync(_compose(_join('\n'), _map(_prop('code')))(arr));
  }

  const pickedSnippet = _map(_pick(keysByLayout))(arr);

  return _compose(console.log, json ? JSON.stringify : colorizedPrint)(pickedSnippet);
};

var version = "1.4.48";

const isTest = process.env.NODE_ENV === 'test';
const actions = {
  random: opts => printSnippet(opts, randomSnippet(snippets)),
  search: (opts, query) => printSnippet(opts, searchSnippets(snippets, query)),
  tag: (opts, id) => printSnippet(opts, getSnippetsByTag(snippets, id)),
  view: (opts, id) => printSnippet(opts, getSnippet(snippets, id))
};

const addCommand = settings => _reduce((acc, [key, ...args]) => acc[key](...args), program, settings);

const addAction = action => ['action', (...args) => ((input, opts = {}) => {
  if (!input) return program.outputHelp();
  return isTest ? // NOTE: Commander.js sadly does not include a way to hook onto given
  // actions for integration testing. We solve this by outputting state
  // when NODE_ENV is test.
  console.log(JSON.stringify([action, input, !!opts.cp, !!opts.json, opts.layout])) : actions[action](opts, input);
})(...(action === 'random' ? [true, args[0]] : args))];

const commonOptions = [['option', '-c, --cp', 'copy code to clipboard', false], ['option', '-j, --json', 'output in json format', false], ['option', '-l, --layout <layout>', 'print in specified layout', 'itced']];
program.version(version);
addCommand([['command', 'r'], ['alias', 'random'], ['description', ['view random snippet']], ...commonOptions, addAction('random')]);
addCommand([['command', 's [query]'], ['alias', 'search'], ['description', ['fuzzy search snippets by id']], ...commonOptions, addAction('search')]);
addCommand([['command', 't [id]'], ['alias', 'tag'], ['description', ['view snippets by tag']], ...commonOptions, addAction('tag')]);
addCommand([['command', 'v [id]'], ['alias', 'view'], ['description', ['view snippet with id']], ...commonOptions, addAction('view')]);
addCommand([['command', '* [id]'], ['description', ['view snippet with id']], addAction('view')]);
program.on('--help', () => console.log(['', 'Examples:', '  30s head', '  30s v head', '  30s view head', '', '  30s r', '  30s random', '', '  30s s -j flatten', '  30s search --json flatten', '', '  30s t -l ic array', '  30s tag --layout ic array'].join('\n')));
program.parse(process.argv);

if (process.argv.length <= 2) {
  program.outputHelp();
}
